@@ -5,13 +5,13 @@ from itertools import izip
 
 class Describer(storytext.guishared.Describer):
     ignoreWidgets = [ swing.JSplitPane, swing.CellRendererPane, swing.Box.Filler, swing.JRootPane, swing.JLayeredPane,
-                      swing.JPanel, swing.JOptionPane, swing.JViewport, swing.table.JTableHeader, swing.JScrollPane,
+                      swing.JPanel, swing.JOptionPane, swing.JViewport, swing.table.JTableHeader, swing.JScrollPane, swing.JFileChooser,
                       swing.JScrollBar, swing.plaf.basic.BasicInternalFrameTitlePane, swing.JInternalFrame.JDesktopIcon ]
     ignoreChildren = (swing.JScrollBar, swing.JMenu, swing.JPopupMenu, swing.JMenuBar, swing.JSpinner, swing.JComboBox,
                       swing.JInternalFrame, swing.plaf.basic.BasicInternalFrameTitlePane)
     statelessWidgets = [ swing.plaf.basic.BasicSplitPaneDivider, swing.JInternalFrame, swing.JSeparator ]
     stateWidgets = [ swing.JButton, swing.JFrame, swing.JMenuBar, swing.JMenu, swing.JMenuItem, swing.JToolBar,
-                    swing.JRadioButton, swing.JCheckBox, swing.JTabbedPane, swing.JDialog, swing.JLabel,
+                    swing.JRadioButton, swing.JCheckBox, swing.JToggleButton, swing.JTabbedPane, swing.JDialog, swing.JLabel,
                     swing.JList, swing.JTree, swing.JTable, swing.text.JTextComponent, swing.JPopupMenu,
                      swing.JProgressBar, swing.JSpinner, swing.JComboBox ]
     childrenMethodName = "getComponents"
@@ -163,7 +163,7 @@ class Describer(storytext.guishared.Describer):
                 grid.append([ "" ])
             while len(grid[y]) <= x:
                 grid[y].append("")
-            grid[y][x] = str(desc)
+            grid[y][x] = unicode(desc)
         return grid, max((len(r) for r in grid))
 
     def getHorizontalSpans(self, children, columnCount):
@@ -259,6 +259,13 @@ class Describer(storytext.guishared.Describer):
     def getJCheckBoxState(self, widget):
         text = "Check box: " + util.ComponentTextFinder(widget, describe=True).getJCheckBoxText()
         return self.combineElements([ text ] + self.getPropertyElements(widget))
+
+    def getJToggleButtonDescription(self, widget):
+        return self.getAndStoreState(widget)
+    
+    def getJToggleButtonState(self, widget):
+        text = "Toggle button: " + util.ComponentTextFinder(widget, describe=True).getJToggleButtonText()
+        return self.combineElements([ text ] + self.getPropertyElements(widget))
         
     def getJTabbedPaneDescription(self, widget):
         state = self.getState(widget)
@@ -462,8 +469,14 @@ class Describer(storytext.guishared.Describer):
 
     def getJComboBoxState(self, widget):
         if widget.isEditable():
-            return str(widget.getEditor().getItem()), self.getPropertyElements(widget)
-        return str(widget.getSelectedItem()), self.getPropertyElements(widget)
+            return self.itemToString(widget.getEditor().getItem()), self.getPropertyElements(widget)
+        return self.itemToString(widget.getSelectedItem()), self.getPropertyElements(widget)
+        
+    def itemToString(self, item):
+        if hasattr(item, "getDescription"):
+            return item.getDescription()
+        else:
+            return str(item)
         
     def getJComboBoxDescription(self, widget):
         contents, properties = self.getJComboBoxState(widget)
