@@ -78,8 +78,8 @@ class UseCaseReplayer(storytext.guishared.ThreadedUseCaseReplayer):
         describer = self.getDescriberClass()()
         runOnUIThread(describer.addFilters, monitor.getDisplay())
         def describe():
-            runOnUIThread(describer.describeWithUpdates, monitor.getActiveShell())
-        self.describeAndRun(describe)
+            runOnUIThread(describer.describeWithUpdates, monitor.getActiveShell)
+        self.describeAndRun(describe, monitor.handleReplayFailure)
 
     def getDescriberClass(self):
         try:
