@@ -65,37 +65,66 @@ class RecordScript:
 class ShortcutTracker:
     def __init__(self, replayScript):
         self.replayScript = replayScript
-        self.unmatchedCommands = []
+        self.commandsForMatch = []
+        self.commandsForMismatch = []
+        self.logger = logging.getLogger("Shortcut Tracker")
         self.reset()
 
     def reset(self):
-        self.replayScript = ReplayScript(self.replayScript.name)
+        self.replayScript = ReplayScript(self.replayScript.name, ignoreComments=True)
+        self.commandsForMatch = copy(self.commandsForMismatch)
         self.argsUsed = []
         self.currRegexp = self.replayScript.getCommandRegexp()
 
+    def hasStarted(self):
+        return self.commandsForMismatch != self.commandsForMatch
+
     def updateCompletes(self, line):
         if self.currRegexp is None:
+            self.logger.debug("Ignore " +  self.replayScript.getShortcutName())
             return False # We already reached the end and should forever be ignored...
         match = self.currRegexp.match(line)
         if match:
+            self.commandsForMismatch.append(line)
+            self.logger.debug("Match " +  self.replayScript.getShortcutName() + ", " + self.currRegexp.pattern \
+                               + ", " +  repr(line) + ", " + repr(self.commandsForMismatch) + ", " + repr(self.commandsForMatch))
             self.currRegexp = self.replayScript.getCommandRegexp()
-            self.argsUsed += match.groups()
+            groupdict = match.groupdict()
+            if groupdict: # numbered arguments
+                for key, val in groupdict.items():
+                    argPos = int(key.replace("var", "")) - 1
+                    while len(self.argsUsed) <= argPos:
+                        self.argsUsed.append("")
+                    self.argsUsed[argPos] = val
+            else:
+                self.argsUsed += match.groups()
             return not self.currRegexp
         else:
-            self.unmatchedCommands += self.replayScript.getCommandsSoFar()
-            self.unmatchedCommands.append(line)
-            self.reset()
-            return False
+            if self.hasStarted():
+                self.reset()
+                self.logger.debug("Reset " + self.replayScript.getShortcutName() + ", " + repr(self.commandsForMismatch))
+                return self.updateCompletes(line)
+            else:
+                self.commandsForMismatch.append(line)
+                self.reset()
+                return False
         
     def rerecord(self, newCommands):
         # Some other tracker has completed, include it in our unmatched commands...
-        self.unmatchedCommands = copy(newCommands)
-
+        started = self.hasStarted()
+        self.commandsForMismatch = copy(newCommands)
+        if self.currRegexp is None: # we completed, but haven't been chosen, because there was a better one to use
+            self.reset()
+        elif not started:
+            self.commandsForMatch = copy(self.commandsForMismatch)
+        self.logger.debug("Rerecord " + self.replayScript.getShortcutName() + ", " + repr(self.commandsForMismatch) + ", " + repr(self.commandsForMatch))
+        
     def getNewCommands(self):
         shortcutName = self.replayScript.getShortcutNameWithArgs(self.argsUsed)
+        self.commandsForMatch.append(shortcutName)
+        self.commandsForMismatch = self.commandsForMatch
         self.reset()
-        self.unmatchedCommands.append(shortcutName)
-        return self.unmatchedCommands
+        return self.commandsForMatch
     
     def isLongerThan(self, otherTracker):
         return len(self.replayScript.commands) > len(otherTracker.replayScript.commands)
@@ -107,6 +136,7 @@ class UseCaseRecorder:
         # Store events we don't record at the top level, usually controls on recording...
         self.eventsBlockedTopLevel = []
         self.scripts = []
+        self.comments = []
         self.processId = os.getpid()
         self.applicationEvents = OrderedDict()
         self.supercededAppEventCategories = {}
@@ -114,7 +144,7 @@ class UseCaseRecorder:
         self.realSignalHandlers = {}
         self.origSignal = signal.signal
         self.signalNames = {}
-        self.stateChangeEventInfo = None
+        self.stateChangeEventInfo = {}
         self.delayedEvents = []
         self.hasAutoRecordings = False
         recordScript = os.getenv("USECASE_RECORD_SCRIPT")
@@ -134,10 +164,19 @@ class UseCaseRecorder:
     def isActive(self):
         return len(self.scripts) > 0
 
+    def storeComment(self, comment):
+        self.logger.debug("Storing comment " + repr(comment))
+        self.comments.append(comment)
+
     def addScript(self, scriptName):
         self.scripts.append(RecordScript(scriptName))
 
-    def closeScripts(self):
+    def closeScripts(self, exitHook):
+        if any((c is not None for c in self.comments)):
+            if exitHook:
+                sys.stderr.write("NOTE: discarded terminal comments in usecase file, these are not supported for Java applications.\n")
+            else:
+                self.recordComments()
         for script in self.scripts:
             script.close()
     
@@ -203,38 +242,51 @@ class UseCaseRecorder:
         self.logger.debug("Event of type " + event.__class__.__name__ + " for recording")
         if not event.shouldRecord(*args):
             self.logger.debug("Told we should not record it : args were " + repr(args))
-            if event.checkPreviousWhenRejected() and self.stateChangeEventInfo:
-                stateChangeOutput, stateChangeEvent, _, _ = self.stateChangeEventInfo
-                if event.implies(stateChangeOutput, stateChangeEvent, *args):
-                    self.logger.debug("Discarded event implies previous state change event, ignoring previous also")
-                    self.stateChangeEventInfo = None
+            if event.checkPreviousWhenRejected():
+                delayLevel = event.delayLevel(*args)
+                if delayLevel in self.stateChangeEventInfo:
+                    stateChangeOutput, stateChangeEvent, _ = self.stateChangeEventInfo[delayLevel]
+                    if event.implies(stateChangeOutput, stateChangeEvent, *args):
+                        self.logger.debug("Discarded event implies previous state change event, ignoring previous also")
+                        del self.stateChangeEventInfo[delayLevel]
             return
         
-        impliesPrevious, writtenAppEvents = False, False
-        if self.stateChangeEventInfo:
-            stateChangeOutput, stateChangeEvent, stateChangeDelayLevel, appEvents = self.stateChangeEventInfo
-            if stateChangeDelayLevel >= event.delayLevel(*args):
-                impliesPrevious = event.implies(stateChangeOutput, stateChangeEvent, *args)
-                writtenAppEvents = self.writeApplicationEventDetails(appEvents, stateChangeDelayLevel)
-                if impliesPrevious:
-                    self.logger.debug("Implies previous state change event, ignoring previous")
-                else:
-                    self.recordOrDelay(stateChangeOutput, stateChangeDelayLevel, stateChangeEvent)
-                self.stateChangeEventInfo = None
+        impliesPrevious = False
+        delayLevel = event.delayLevel(*args)
+        if delayLevel in self.stateChangeEventInfo:
+            stateChangeOutput, stateChangeEvent, appEvents = self.stateChangeEventInfo[delayLevel]
+            impliesPrevious = event.implies(stateChangeOutput, stateChangeEvent, *args)
+            self.writeApplicationEventDetails(appEvents, delayLevel)
+            if impliesPrevious:
+                self.logger.debug("Implies previous state change event, ignoring previous")
+            else:
+                self.recordOrDelay(stateChangeOutput, delayLevel, stateChangeEvent)
+            del self.stateChangeEventInfo[delayLevel]
 
         scriptOutput = event.outputForScript(*args)
-        delayLevel = event.delayLevel(*args)
         if event.isStateChange() and delayLevel >= self.getMaximumStoredDelay():
-            self.logger.debug("Storing up state change event " + repr(scriptOutput) + " with delay level " + repr(delayLevel))
-            appEvents = {} if impliesPrevious else copy(self.applicationEvents)
-            self.stateChangeEventInfo = scriptOutput, event, delayLevel, appEvents
+            appEvents = {} if impliesPrevious else self.transferAppEvents(delayLevel)                
+            self.logger.debug("Storing up state change event " + repr(scriptOutput) + " with delay level " + repr(delayLevel) + " and app events " + repr(appEvents))
+            self.stateChangeEventInfo[delayLevel] = scriptOutput, event, appEvents
         else:
-            if not writtenAppEvents or not impliesPrevious:
-                self.writeApplicationEventDetails(self.applicationEvents, delayLevel)
+            # If impliesPrevious is true, it means any app events since the last state change event were essentially generated by the current event
+            # In that case they should not be recorded at the moment. We still need to adjust the delay levels though
+            delayLevelForAppEvents = max(delayLevel, int(impliesPrevious))
+            self.writeApplicationEventDetails(self.applicationEvents, delayLevelForAppEvents)
             if self.recordOrDelay(scriptOutput, delayLevel, event):
                 self.processDelayedEvents(self.delayedEvents)
                 self.delayedEvents = []
 
+    def transferAppEvents(self, delayLevel):
+        appEvents = OrderedDict()
+        for appEventKey, eventName in self.applicationEvents.items():
+            _, currDelayLevel = appEventKey
+            if currDelayLevel < delayLevel:
+                continue
+            appEvents[appEventKey] = eventName
+            del self.applicationEvents[appEventKey]
+        return appEvents
+
     def getMaximumStoredDelay(self):
         return max((i[1] for i in self.delayedEvents)) if self.delayedEvents else 0
 
@@ -275,8 +327,21 @@ class UseCaseRecorder:
                             nextLevelEvents = []
                 else:
                     nextLevelEvents.append((scriptOutput, delayLevel, source))
+        
+    def recordComments(self):
+        while len(self.comments) > 0:
+            comment = self.comments.pop(0)
+            if comment is not None:
+                self._record(comment)
+            else:
+                self.logger.debug("Got None in comment list, not recording more comments now")
+                break
                 
     def record(self, line, event=None):
+        self.recordComments()
+        self._record(line, event)
+        
+    def _record(self, line, event=None):
         self.logger.debug("Recording " + repr(line))
         self.hasAutoRecordings |= line.startswith("Auto.")
         for script in self.getScriptsToRecord(event):
@@ -332,12 +397,13 @@ class UseCaseRecorder:
                 categories.add(newCategory)
                 self.logger.debug("Swapping for " + repr(supercedeCategory) + ": " + repr(oldCategory) + " -> " + repr(newCategory))
 
-    def applicationEventDelay(self, name):
+    def applicationEventDelay(self, name, fromLevel=0, increase=True):
         for appEventKey, eventName in self.applicationEvents.items():
             categoryName, oldDelayLevel = appEventKey
-            if eventName == name and oldDelayLevel == 0:
+            if eventName == name and oldDelayLevel == fromLevel:
                 del self.applicationEvents[appEventKey]
-                self.registerApplicationEvent(name, categoryName, delayLevel=1)
+                newDelayLevel = oldDelayLevel + 1 if increase else oldDelayLevel - 1
+                self.registerApplicationEvent(name, categoryName, delayLevel=newDelayLevel)
                 
     def makeMultiple(self, text, count):
         if count == 1:
@@ -365,16 +431,7 @@ class UseCaseRecorder:
             if currDelayLevel < minDelayLevel:
                 continue
             appEventInfo.setdefault(currDelayLevel, []).append((eventName, categoryName))
-            if appEventKey in self.applicationEvents:
-                storedFullName = self.applicationEvents[appEventKey]
-                if eventName == storedFullName:
-                    del self.applicationEvents[appEventKey]
-                else:
-                    storedName, storedCount = self.parseMultiples(storedFullName)
-                    givenName, givenCount = self.parseMultiples(eventName)
-                    if storedName == givenName:
-                        remainder = storedCount - givenCount
-                        self.reduceApplicationEventCount(givenName, categoryName, currDelayLevel, remainder)
+            del events[appEventKey]
         return appEventInfo
                             
     def writeApplicationEventDetails(self, events, minDelayLevel):
@@ -383,7 +440,6 @@ class UseCaseRecorder:
             eventNames = sorted((e[0] for e in eventInfo))
             eventString = ", ".join(eventNames)
             self.recordOrDelay(waitCommandName + " " + eventString, delayLevel, eventInfo)
-        return len(appEventInfo) > 0
 
     def registerShortcut(self, replayScript):
         for script in self.scripts:
